"""
Adopted from jenkins sources: core/src/main/java/hudson/util/Secret.java
Inspired by:

- https://stackoverflow.com/questions/25547381/what-password-encryption-jenkins-is-using/47020084#47020084  # NOQA
- https://github.com/tweksteen/jenkins-decrypt
"""
import re
import os
import hashlib
from struct import unpack, pack

from ... compat import AES

#: default jenkins magic
MAGIC = b'::::MAGIC::::'
#: by default
BLOCK_SIZE = DEFAULT_IV_LENGTH = 16
#: appears to be a random seed, but jenkins knows a drill
IV = b'\x00' * BLOCK_SIZE

#: CBC type
JENKINS_CIPHER_CBC_TYPE = b'\x01'


def new_iv(length: int = DEFAULT_IV_LENGTH) -> bytes:
    """
    generates new initialization vector
    """
    return os.urandom(length)


def pad(text: bytes, block_size=BLOCK_SIZE):
    size = (block_size - len(text) % block_size)
    return text + (chr(size) * size).encode('utf-8')


class Secret(object):
    """
    Jenkins Cipher provides encryption / decryption operations for:

    - AES.MODE_ECB (before security-304)
    - AES.MODE_CBC (after applying patches security-304)

    See https://jenkins.io/security/advisory/2017-02-01/ for more details
    """
    def __init__(self, master_key: bytes, hudson_secret_key: bytes):
        """
        Initiate jenkins cipher object with master and hudson secret keys
        for further encryption / decryption procedures

        :param master_key: jenkins master key
        :param hudson_secret_key: hudson secret key
        :raises AssertionError:
            - if MAGIC was not found in processed data
        """
        hashed_master_key = hashlib.sha256(master_key).digest()[:BLOCK_SIZE]
        cipher = AES.new(hashed_master_key, AES.MODE_ECB)
        result = cipher.decrypt(hudson_secret_key)
        assert MAGIC in result

        key = result[:-16]
        self.cipher_key = key[:16]

    @staticmethod
    def _is_payload_v1(raw) -> bool:
        """
        Detects if data has been encrypted with new approach

        :param bytes raw: encrypted string
        :return: True if data was encrypted with new approach
        """
        return raw[0] == 1

    def _decrypt_cbc(self, encrypted_text: bytes) -> str:
        """
        Decrypt data using AES.MODE_CBC

        encrypted text size::

            data {
                encryption type: byte, // byte
                iv_size: byte * 4, // int
                data_size: byte * 4, // int
                iv: byte * iv_size, // variable size
                block: byte * data_size // variable size
            }

        :param bytes encrypted_text: encrypted text to decipher
        :return: decrypted text based on cipher block chaining data
        """
        block = encrypted_text[1:]
        iv_size, *_ = unpack('>I', block[:4])

        #: strip iv length int block
        block = block[4:]
        #: using big-endian
        data_size, *_ = unpack('>I', block[:4])

        #: strip data length int block
        block = block[4:]
        iv = block[:iv_size]
        block = block[iv_size:]
        decrypter = AES.new(self.cipher_key, AES.MODE_CBC, iv)
        decrypted = decrypter.decrypt(block)
        padding = decrypted[-1]
        return decrypted[:len(decrypted) - padding].decode('utf-8')

    def _decrypt_ecb(self, encrypted_text: bytes) -> str:
        """
        Decrypt AES.MODE_ECB encrypted content (jenkins versions
        before security-304)

        :param encrypted_text:
        :return: decrypted text
        :raises AssertionError:
            - if MAGIC was not found in decrypted data taken from
              ``encrypted_text``
        """
        cipher = AES.new(self.cipher_key, AES.MODE_ECB)
        raw = cipher.decrypt(encrypted_text)
        assert MAGIC in raw
        return re.split(MAGIC.decode('utf-8'), raw.decode('utf-8'), 1)[0]

    def decrypt(self, encrypted_text: bytes) -> str:
        """
        Decrypts jenkins ``password_hash`` given in base64 encoded string to
        plain password, automatically detects old (ECB) and new (CBC)
        encrypted passwords.

        :param encrypted_text: base64 encoded password hash
            .. note::

                password hash should be in bytes form, not in base64 encoded
                string format.

        :return: plain password
        """
        if self._is_payload_v1(encrypted_text):
            return self._decrypt_cbc(encrypted_text)
        return self._decrypt_ecb(encrypted_text)

    def _encrypt_cbc(self, plain_text: str) -> bytes:
        """
        Encrypts text using AES.MODE_CBC
        (since https://jenkins.io/security/advisory/2017-02-01/)

        :param plain_text: text to encrypt
        :return: encrypted text
        """
        iv = new_iv()
        iv_size = pack('>I', len(iv))

        encrypter = AES.new(self.cipher_key, AES.MODE_CBC, iv)
        encrypted = encrypter.encrypt(
            pad(plain_text.encode('utf-8'))
        )
        encrypted_size = pack('>I', len(encrypted))

        payload = (
            JENKINS_CIPHER_CBC_TYPE + iv_size + encrypted_size +
            iv + encrypted
        )
        return payload

    def encrypt(self, plain_text: str, cipher_type: int=AES.MODE_ECB) -> bytes:
        """
        Encrypt ``plain`` string with jenkins hudson secret key

        :param plain_text: plain text to encrypt
        :param cipher_type: cipher type, one of the following:

        - AES.MODE_ECB
        - AES.MODE_CBC

        :return: encrypted text
        :raises NotImplementedError:

            - if cipher type is not supported yet

        .. note::

            encrypted text are not encoded to base64, so please pack it after
        """
        if cipher_type == AES.MODE_ECB:
            cipher = AES.new(self.cipher_key, AES.MODE_ECB)
            message = pad(plain_text.encode('utf-8') + MAGIC)
            return cipher.encrypt(message)
        elif cipher_type == AES.MODE_CBC:
            return self._encrypt_cbc(plain_text)
        else:
            raise NotImplementedError("Not implemented")



import base64
import random
import string
import unittest

from jenkins import Secret
from jenkins.compat import AES


def fuzzy_text(length: int=12) -> str:
    return (
        ''.join([random.choice(string.ascii_letters + string.digits)
                 for _ in range(length)])
    )


class JenkinsTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        master_key = (
            b"12780d3c3d93e23360dbe0e8750a9a1ba1c521baec776e956dc"
            b"214f32ade1f27a4956490a9152014d2944ab8b82261f17858926"
            b"5dd54ecd659eca965dbbf537c2aa08e10418de19bb96f4c44ed9"
            b"c42f702f7ca2fce790730dfad7f99cf0debc194508f3cde4baf6"
            b"01cc64f7efd0377259bcafa1ee9901efc900eb6a216740757"
        )
        hudson_secret_key = (
            b"\x65\x02\xbf\x0d\xa4\x37\x22\x6b\xb0\x40\xab\x19\x9c\xf1\xb2\x1b"
            b"\x6c\x3f\xa1\x5b\xab\x2f\x0b\x0f\xcc\x56\x36\xb6\xa0\x8d\x35\x98"
            b"\xc5\x14\x8a\xd9\x99\x53\xb6\x98\x59\x2f\x90\x81\x1c\x53\xc9\xe1"
            b"\x61\xc7\x61\x18\x1a\x06\x69\x49\x91\xf7\x59\x7d\xcd\xce\x07\x2e"
            b"\xd3\x9f\xee\xc3\x13\x18\x72\x7e\x47\x8f\xfe\x77\x9a\x95\x75\xf3"
            b"\x9b\x38\xc6\xde\xcd\x27\xf4\x95\x00\xd8\xac\x69\x3e\xfd\x04\xb1"
            b"\x1a\x41\x93\xe5\x27\xdd\xb7\x02\x28\xf0\x0c\x03\xa5\xa8\x69\x2d"
            b"\xeb\x57\x86\x77\x69\x37\xda\x5e\xf4\xdb\xdb\x1a\x28\x3b\x43\xef"
            b"\x56\x3d\x51\xf7\xa3\x1d\xd9\x04\xef\x77\x72\x1a\x8a\xd9\x31\x2f"
            b"\x11\xa8\x49\x83\xee\xbc\x8a\x40\x9d\xa4\xdf\x90\xd0\xa9\x36\x89"
            b"\x7b\x02\x4a\xeb\x83\x3c\xc9\xb9\x5f\xe1\x53\x26\xe9\x62\xe6\xe4"
            b"\x47\xa5\xdd\x3a\xc4\xda\x3b\xf5\x40\xd7\x13\xff\x2f\x50\x3e\x0b"
            b"\x3b\x16\xaa\x32\x1d\xdb\xd9\xec\x7e\x23\x22\x15\x23\x71\x87\xc6"
            b"\x61\x8a\xe6\xf6\xcf\xd3\x9c\xcb\x62\x9c\xa7\xb6\xf6\x98\xbf\x33"
            b"\x1b\x92\xa5\xf7\x04\xf4\x98\x3e\xfb\x42\x1a\x17\xd9\xb0\xcc\xe5"
            b"\xec\x7c\x8d\xd6\x7b\x17\xd3\xa5\x15\xa0\xb3\x69\x16\xdc\xbe\x3d"
            b"\x73\x77\x55\xe1\x45\x7d\x8b\x54\x3d\xe5\xfd\x0d\xe9\xb0\x8b\x71"
        )
        #: "jenkins" word is encrypted
        cls.ecb_password_encrypted = base64.decodebytes(
            b"S3utyMSH+P2IcVQNFtXTDKS3SJa9ZTsGOqquMmggSdA="
        )

        #: "jenkins" word is encrypted
        cls.cbc_password_encrypted = base64.decodebytes(
            b"{AQAAABAAAAAQEbSJ/dEihj3MfVrukQTcsGbkUSNjeYlStrHiooNyY6c=}"
        )
        cls.dummy_password_plain = 'jenkins'
        cls.secret = Secret(
            master_key=master_key,  hudson_secret_key=hudson_secret_key
        )

    def test_decrypt_ecb(self):
        self.assertEqual(
            self.secret.decrypt(self.ecb_password_encrypted),
            self.dummy_password_plain
        )

    def test_encrypt_ecb(self):
        encrypted_password = self.secret.encrypt(
            self.dummy_password_plain
        )
        self.assertEqual(self.ecb_password_encrypted, encrypted_password)
        self.assertEqual(
            self.secret.decrypt(encrypted_password),
            self.dummy_password_plain
        )

    def test_encrypt_ecb_fuzzy_text(self):
        text = fuzzy_text(19)
        encrypted_password = self.secret.encrypt(
            text, cipher_type=AES.MODE_ECB
        )
        self.assertEqual(self.secret.decrypt(encrypted_password), text)

    def test_encrypt_cbc_fuzzy_text(self):
        text = fuzzy_text(19)
        encrypted_password = self.secret.encrypt(
            text, cipher_type=AES.MODE_CBC
        )
        self.assertEqual(self.secret.decrypt(encrypted_password), text)

    def test_decrypt_cbc(self):
        encrypted = base64.decodebytes(
            b"{AQAAABAAAAAg2nYSYPQRMrY5PZxLLuble6jVKmd7vn"
            b"VN9P2eCpUf6yi9HomkwaSQu8MKG5gBrgb0}"
        )
        self.assertEqual(
            self.secret.decrypt(encrypted), "test encrypted text"
        )

    def test_decrypt_unsupported(self):
        with self.assertRaises(NotImplementedError):
            self.secret.encrypt("test", cipher_type=AES.MODE_CFB)



import os
import sys
import base64
import argparse

import logging

from . import Secret
from ... compat import AES


logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


ACTION_ENCRYPT = 'encrypt'
ACTION_DECRYPT = 'decrypt'
ACTIONS = (
    ACTION_ENCRYPT, ACTION_DECRYPT
)
DEFAULT_ACTION = ACTION_ENCRYPT
EXIT_OK = 0

ENC_ECB = 'ecb'
ENC_CBC = 'cbc'
ENC_TYPES = (ENC_ECB, ENC_CBC)
ENC_TYPE_MAP = {
    ENC_ECB: AES.MODE_ECB,
    ENC_CBC: AES.MODE_CBC
}


def get_absolute_path(relative_path, lower_case=True):
    """
    gets absolute path in respect of running operating system (useful to use
    it for volumes registration and so other basic operations)

    :param str relative_path: relative path to get absolute path for docker
        volume. Depending on OS it could be different.
    :param bool lower_case: cast lower case if True (default)
    :rtype: str
    :return: str
    """
    path = os.path.abspath(relative_path).replace('\\', '/')
    if lower_case:
        return path.lower()
    return path


def main(opts):
    master_key_location = get_absolute_path(opts.master_key)
    hudson_secret_key_location = get_absolute_path(opts.hudson_secret_key)
    master_key = open(master_key_location, 'rb').read()
    hudson_secret_key = open(hudson_secret_key_location, 'rb').read()
    cipher = Secret(hudson_secret_key=hudson_secret_key,
                    master_key=master_key)
    for message in opts.messages:
        if opts.action == ACTION_ENCRYPT:
            cipher_type = ENC_TYPE_MAP.get(opts.aes_type,
                                           ENC_TYPE_MAP[ENC_ECB])
            encrypted = cipher.encrypt(message, cipher_type=cipher_type)
            b64_encoded = base64.b64encode(encrypted).decode('utf-8')
            if opts.aes_type == ENC_CBC:
                b64_encoded = '{%s}' % b64_encoded
            print(b64_encoded)
        elif opts.action == ACTION_DECRYPT:
            print(cipher.decrypt(base64.b64decode(message)))
        else:
            pass
    return EXIT_OK


def execute():
    parser = argparse.ArgumentParser(
        prog='jenkins_cipher',
        description='cipher operations (encrypt / decrypt)'
    )
    parser.add_argument(nargs='*', dest='messages', type=str,
                        help='passwords split by space to encrypt',
                        metavar='text')
    parser.add_argument(
        '--action', dest='action',
        choices=ACTIONS,
        default=DEFAULT_ACTION,
        help='action to perform, default is `%s`' % DEFAULT_ACTION,
        required=False
    )
    parser.add_argument(
        '--aes-type', dest='aes_type',
        default=ENC_ECB,
        choices=ENC_TYPES,
        help='encryption type, default: %(default)s, '
             'available choices: %(choices)s '
             'Jenkins before 2017-02-01 uses ECB, newer jenkins uses CBC, '
             'see https://jenkins.io/security/advisory/2017-02-01/ for details'
    )
    parser.add_argument('--master-key', dest='master_key',
                        required=True,
                        help='master key location')
    parser.add_argument('--hudson-secret-key', dest='hudson_secret_key',
                        help='hudson secret key')
    options = parser.parse_args()
    sys.exit(main(options))


if __name__ == '__main__':
    execute()



import os
from setuptools import find_packages, setup

PACKAGE_ROOT = os.path.abspath(os.path.dirname(__file__))


def relative_path(path, base_dir=PACKAGE_ROOT):
    return os.path.join(base_dir, path)


requirements = [
    'pycryptodomex>=3.6.1;platform_system=="Windows"',
    'pycrypto>=2.6.1;platform_system != "Windows"'
]

classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Other Environment',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    # 'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Software Development :: Libraries',
]

test_requirements = ['tox>=2.9.1']
setup_requirements = ['setuptools-scm']


def readme():
    with open(relative_path('README.rst'), 'r') as f:
        return f.read()


setup(
    author='Nickolas Fox',
    author_email='tarvitz@blacklibrary.ru',
    url="https://github.com/tarvitz/jenkins-secret",
    name='jenkins-utils',
    install_requires=requirements,
    tests_require=test_requirements,
    setup_requires=setup_requirements,
    classifiers=classifiers,
    description='Jenkins Utils',
    long_description=readme(),
    keywords='jenkins cipher secret tools utils aes ecb cbc',
    python_requires='>=3',
    packages=find_packages(),
    platforms=['any'],
    include_package_data=True,
    zip_safe=False,
    use_scm_version={
         'write_to': 'jenkins/_version.py',
    },
    entry_points={
        'console_scripts': [
            'jenkins-secret = jenkins.utils.secret.__main__:execute',
        ]
    },
)



#!/usr/bin/env python3
"""
.. synopsis: Development invoke script, helps with debug purpose
.. author:: Nickolas Fox <tarvitz@blacklibrary.ru>
"""

import sys
from jenkins.utils.secret import __main__


def check():
    if sys.version_info[:2] < (3, 4):
        print("jenkins-cipher requires python-3.4+, exiting ..")
        sys.exit(-1)


def main():
    sys.exit(__main__.execute())


if __name__ == '__main__':
    main()



# compat module for crypto functions
try:
    from Crypto.Cipher import AES
except ImportError:
    try:
        from Cryptodome.Cipher import AES
    except ImportError as imp_err:
        raise imp_err

__all__ = ['AES']
